#version 330 core

#define WATER_COLOR vec4(0.23, 0.35, 0.38, 0.95)

uniform sampler2D terrainTex;
uniform sampler2D depthTex;

uniform vec2 blockTextureSize;
uniform vec2 textureAtlasSize;
uniform float waterFlowCounter;
uniform vec2 frameSize;
uniform vec3 sunPosRel;
uniform vec3 sunColor;
uniform vec3 reflectColor;
uniform float waterWaveCounter;
uniform float sunSpecularIntensity;
uniform float dropletIntensity = 0;

in vec4 rgba;
in vec4 rgbaFog;
in float fogAmount;
in vec2 uv;
in vec2 uvSize;
in float waterStillCounterOff;
flat in vec2 uvBase;
in vec3 fragWorldPos;
in vec3 fWorldPos;
in vec3 fragNormal;
in float fresnel;
flat in int skyExposed;

in vec2 flowVectorf;
in float glowLevel;

flat in int waterFlags;
flat in int renderFoam;

layout(location = 0) out vec4 outAccu;
layout(location = 1) out vec4 outReveal;
layout(location = 2) out vec4 outGlow;

#include fogandlight.fsh
#include noise2d.ash
#include noise3d.ash
#include colormap.fsh

// https://github.com/hughsk/glsl-luma/blob/master/index.glsl
float luma(vec3 color) {
  return dot(color, vec3(0.299, 0.587, 0.114));
}

vec2 droplethash3( vec2 p )
{
    vec2 q = vec2(dot(p,vec2(127.1,311.7)), dot(p,vec2(269.5,183.3)));
    return fract(sin(q)*43758.5453);
}

float dropletnoise(in vec2 x)
{
    if (dropletIntensity < 0.001) return 0.;
	
    x *= dropletIntensity;
    
    vec2 p = floor(x);
    vec2 f = fract(x);


    float va = 0.0;
    for( int j=-1; j<=1; j++ )
    for( int i=-1; i<=1; i++ )
    {
        vec2 g = vec2(float(i), float(j));
        vec2 o = droplethash3(p + g);
        vec2 r = g - f + o;
        float d = length(r) / dropletIntensity;
        
        float a = max(cos(d - waterWaveCounter * 2.7 + (o.x + o.y) * 5.0), 0.);
        a = smoothstep(0.99, 0.999, a);
        
        float ripple = mix(a, 0., d);
        va += max(ripple, 0.);
    }
	
    return va;
}





void drawPixel(vec4 color, float accuWeight, float revealStrength, float sparkleGlowAmount) {
	float weight = color.a * clamp(0.03 / (1e-5 + pow(gl_FragCoord.z / 200, 4.0)), 1e-2, 3e3);
	
    // RGBA32F texture (accumulation)
    outAccu = vec4(color.rgb * color.a, color.a) * weight * accuWeight;
	
    // R32F texture (revealage)
    // Make sure to use the red channel (and GL_RED target in your texture)
    outReveal.r = min(color.a * revealStrength, 1.0); // Was *1.2 but that made lava more transparent :o
	
    outGlow = vec4(glowLevel + sparkleGlowAmount, 0, 0, color.a);
}

vec4 toGrayscale(in vec4 color)
{
  float average = (color.r + color.g + color.b) / 3.0;
  return vec4(average, average, average, 1.0);
}

vec4 colorize(in vec4 grayscale, in vec4 color)
{
    return (grayscale * color);
}

void main() 
{
	// When looking through tinted glass you can clearly see the edges where we fade to sky color
	// Using this discard seems to completely fix that
	if (rgba.a < 0.005) discard;
	
	vec4 texColor;
	float sparkleGlowAmount = 0.0;
	float diffTotal = 0;

	
	float vn = max(0, 0.9 - abs(fragNormal.y));
	float wfc;
	
	bool isLava = (waterFlags & 0x2000000) > 0;
	
	if (isLava) wfc = waterFlowCounter / 30.0 * (1 + 5 * vn);
	else wfc = waterFlowCounter / 1.8 * (1 + 5 * vn);

	float flowSpeed = length(flowVectorf);
	if (flowSpeed > 0.001) {
		vec2 flowVec = normalize(flowVectorf);
		if (isLava) flowVec /= 8;
		
		if (fragNormal.y < 0) wfc*=-1;
		
		vec2 uvxOffset = 
			clamp(
				mod((uv - uvBase) + flowVec * wfc * blockTextureSize, blockTextureSize),
				vec2(1 / textureAtlasSize), 
				blockTextureSize - 1 / textureAtlasSize)
		;
		
		texColor = texture(terrainTex, uvBase + uvxOffset);
		
	} else {
		// Needs to be rewritten to not do weird uv-inverse math but simply use a second texture so json blocks can use it too
		
		vec2 uvxOffset = 
			clamp(
				blockTextureSize - uvSize,
				vec2(1 / textureAtlasSize), 
				blockTextureSize - 1 / textureAtlasSize)
		;
		
		texColor = texture(terrainTex, uv) * waterStillCounterOff + (1-waterStillCounterOff) * texture(terrainTex, uvBase + uvxOffset);				
	}
	
	texColor = getColorMapped(terrainTex, texColor);
	if (!isLava) // color water
	{
		texColor = colorize(toGrayscale(texColor), WATER_COLOR * vec4(sunColor.rgb, 1.0));
	}

	vec4 rgbaFinal = rgba;
#if FOAMEFFECT > 0
	rgbaFinal.a = rgba.a + max(0, texColor.a - 0.4);
#else
	rgbaFinal.a = rgba.a + texColor.a;
#endif
	float bright = (rgba.r + rgba.g + rgba.b)/3;
	
	float shadowBright = getBrightnessFromShadowMap();
	
	
	float x = gl_FragCoord.x / frameSize.x;
	float y = gl_FragCoord.y / frameSize.y;
	
	// This seems to fix being able to see rivers when looking up from inside a lake
	if (fogAmount > 0.98) discard;

	if (isLava) {
		texColor *= vec4(vec3((rgbaFinal.r + rgbaFinal.g + rgbaFinal.b)/3 * shadowBright), rgbaFinal.a);
	} else {
		texColor *= vec4(rgbaFinal.rgb * shadowBright, rgbaFinal.a);
	}
	if (flowSpeed > 0.001) {
		texColor.a *= 1.2;
	}

	bool doLightFoam = (waterFlags & 0x4000000) != 0;
	
	// Was * 2 but that made water behind quartz glass super visible in the night
	// Was * 0.5 but that made water columns hardly visible
	float accuWeight = 1;
	
#if FOAMEFFECT > 0
	if (rgbaFinal.a > 0) {
	
		// Water edge + shinyness shading effect, kinda nice
		float ownDepth = linearDepth(gl_FragCoord.z);
		diffTotal = 0;
		int range = 2;
		for (int dx = -range; dx <= range; dx++) {
			for (int dy = -range; dy <= range; dy++) {
				float diff = ownDepth - linearDepth(texture(depthTex, vec2(x + dx/frameSize.x, y + dy/frameSize.y)).x);
				if (diff < 0.001) { // This check prevents foam not rendered when looking through grass at distant water
					diffTotal += abs(diff);
				}
			}
		}
		
		diffTotal /= (4*range * range);
		diffTotal = min(diffTotal, -vn/10 + 0.05);
		
		if (isLava) {
			float intensity = clamp(dot(fragNormal, vec3(0, 1, 0)), 0, 1) * 0.5;
			float a = fragWorldPos.x + fragWorldPos.y - 1.5 * flowVectorf.x * wfc;
			float b = fragWorldPos.z - 1.5 * flowVectorf.y * wfc;
			
			float diff = intensity * clamp(1 - diffTotal*1000 - gnoise(vec3(a*35, b*35, wfc))/2 + gnoise(vec3(a*2, b*2, wfc))/2, 0, 1);
			float noise = intensity * (gnoise(vec3(a, b, wfc)) + 0.5) / 2;
			float rgbAdd = bright*(diff * 0.3 + noise/10);
			texColor.rgb -= vec3(rgbAdd, rgbAdd, rgbAdd);
			texColor.a=1; // Let's just do lava as opaque as we can
			accuWeight=1;
			
			float blackSpots = gnoise(fragWorldPos.xyz) + 0.5;
			
			texColor.rgb -= blackSpots * 1.5 * 0.5;
			texColor.g += 0.2 * 0.5;
			
		} else {
			// Cold liquids
			vec3 localPos = fragWorldPos.xyz;
			
			// Foam
			float intensity = clamp(dot(fragNormal, vec3(0, 1, 0)), 0, 1) * 0.025;
			
			float a = localPos.x + localPos.y - 1.5 * flowVectorf.x * wfc;
			float b = localPos.z - 1.5 * flowVectorf.y * wfc;
			
			// without the -abs() one diagonal direction goes derpy noise
			float noise1 = gnoise(vec3(a*15, -abs(b*15), wfc)) + gnoise(vec3(a*5, b*5, wfc));
			float noise2 = gnoise(vec3(a, b, wfc));
			
			float diff = intensity * clamp(1 - diffTotal*1500 - noise1/2 + noise2/2, 0, 1);
			float noise = intensity * (gnoise(vec3(a * 0.5, b * 0.5, wfc))/2 + gnoise(vec3(a, b, wfc))/2 + 0.5) * 2;
			
			float rgbAdd = max(0, bright*(diff * 0.3 + noise/10));
			if (doLightFoam) {
				rgbAdd /= 2;
			}
			
			texColor.rgb += vec3(rgbAdd, rgbAdd, rgbAdd);
			texColor.a += (max(0, diff/16 + noise/4) + vn / 4) / clamp(fresnel, 0.5, 1);

			// water roughness/shinyness
			const float granularity = 15;
			vec3 noisepos = ((noise * noise1 * noise2) + noise2 * 10 + noise1 * 5) * 0.015 + vec3(localPos.x, localPos.z, waterWaveCounter * 0.015 + windWaveCounter * 0.033) * 0.5;
			float dy = cnoise(noisepos * granularity) * pow((1 - gl_FragCoord.z), 0.65);
			vec3 normal = normalize(vec3(dy, 1, -dy));
			float upness = max(0, dot(fragNormal, vec3(0,1,0))); // Only do specular reflections on up faces
			vec3 eye = normalize(vec3(fWorldPos.x, fWorldPos.y - 2, fWorldPos.z));
			vec3 reflectionVec = reflect(sunPosRel, normal);
			float p = dot(reflectionVec, eye);
			float waterLuma = luma(texColor.rgb);
			texColor.rgb += dy * waterLuma * 3;

			// water clarity
			texColor -= vec4(vec3(0), 0.015) * (3 + fWorldPos.y);
			texColor -= vec4(vec3(0), 0.15) * (5 + fWorldPos.y) * (1 - upness);
			
			// Droplet noise
			float f = 0;
			if (skyExposed > 0) {
				vec2 uv = localPos.xz * (12.0 / (2.0 + noise1/5000.0 + noise/600.0));
				f = dropletnoise(uv);
			}

			// Specular reflection
			#if SHADOWQUALITY == 0
			if (skyExposed > 0) {
			#endif
				if (p > 0) {
					// sun reflection
					float sunSpecularAmount = clamp(pow(distance(localPos, sunPosRel), p * 0.5) * 0.005, 0, 0.5) * 10;
					texColor.rgb += sunColor * (waterLuma * sunSpecularAmount) * upness;
					sparkleGlowAmount += sunSpecularAmount * 0.05;

					// water sparkle
					float sparkleNoise = dy * cnoise2(vec2(-wfc * 3 - a * 16, wfc - b * 16)) * cnoise2(vec2(wfc * 2 - a * 64, wfc + b * 64)) * 20;
					float sparkleIntensity = clamp(dot(fragNormal, sunPosRel) * (1 - gl_FragCoord.z) * 50, 0, 1);
					if (sunSpecularAmount > 0.5 && sparkleNoise > 0.5)
					{
						texColor.rgb += sunColor * sparkleNoise * sparkleIntensity * upness;
						sparkleGlowAmount += sparkleNoise * sparkleIntensity * upness;
					}
				}
			#if SHADOWQUALITY == 0
			}
			#endif
			
			texColor.rgb *= 1 + f;
		}
	}
	
#else
	if ((waterFlags & 0x2000000) > 0) { // Bit 25
		texColor.a=1;
		accuWeight=1;		
	}
#endif
    
	
	texColor = applyFog(texColor, fogAmount + diffTotal);
	
	
	drawPixel(texColor, accuWeight, isLava ? 2 : 1.45, sparkleGlowAmount);
}
